package fullstack

import (
	"fmt"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

func TestJobScheduling(t *testing.T) {
	ft := InitFullstack(t, true)
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))

	t.Run("Detached jobs are not resheduled",
		func(t *testing.T) {
			// send two enveloppes
			status, err := emitter.SendItems(nil, api.DefaultOutput, "evt1", []api.Item{[]byte("item1")})
			if err != nil {
				t.Fatal(err)
			}
			assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)
			status, err = emitter.SendItems(nil, api.DefaultOutput, "evt1", []api.Item{[]byte("item2")})
			if err != nil {
				t.Fatal(err)
			}
			assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

			// Get one message and disconnect
			{
				var (
					xsub *xbus.Subscription
					err  error
					done = make(chan struct{})
				)
				xsub, err = consumer.Subscribe(
					func(apc *xbus.ActorProcessingContext) error {
						fmt.Println("Hello there")
						defer func() {
							fmt.Println("done...")
							done <- struct{}{}
						}()
						apc.ActorLeaving()
						envelope, err := apc.CompleteEnvelope("default", time.Millisecond)
						fmt.Println("got the envelope")
						if err != nil {
							return err
						}
						fmt.Println("checking the envelope")
						item := string(envelope.Events[0].Items[0])
						if item != "item1" {
							fmt.Println("oh oh")
							return fmt.Errorf(
								"Invalid item received. Expected 'item1', got '%s'", item)
						}
						apc.Detach()
						fmt.Println("HERE I AM")
						return nil
					})
				if err != nil {
					t.Fatal(err)
				}
				select {
				case <-done:
					xsub.Unsubscribe()
					fmt.Println("1st round ok")
				case <-time.After(time.Second):
					t.Fatal("no process request received")
				}
			}
			// Reconnect for 1 message, it should be the second one
			{
				var (
					xsub *xbus.Subscription
					err  error
					done = make(chan struct{})
				)
				xsub, err = consumer.Subscribe(
					func(apc *xbus.ActorProcessingContext) error {
						fmt.Println("1")
						defer func() {
							done <- struct{}{}
						}()
						apc.ActorLeaving()
						fmt.Println("2")
						envelope, err := apc.CompleteEnvelope("default", time.Millisecond)
						if err != nil {
							return err
						}
						fmt.Println("3")
						item := string(envelope.Events[0].Items[0])
						if item != "item2" {
							fmt.Println("4")
							return fmt.Errorf(
								"Invalid item received. Expected 'item2', got '%s'", item)
						}
						apc.Detach()
						fmt.Println("HERE I AM, again")
						return nil
					})
				if err != nil {
					t.Fatal(err)
				}
				select {
				case <-done:
					fmt.Println("2st round ok")
					xsub.Unsubscribe()
				case <-time.After(time.Second):
					fmt.Println("timeout !!")
					t.Fatal("no process request received")
				}
			}
		})
}
