package process

import (
	"bytes"
	"context"
	"fmt"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/graph"
	"bitbucket.org/orus-io/xbus/storage"
	"gopkg.in/yaml.v2"
)

const envelopeExportChunkSize = 250 * 1024

// EnvelopeActorPair contains a pair (EnvelopeID, ActorID)
type EnvelopeActorPair struct {
	EnvelopeID api.UUID
	ActorID    api.UUID
}

// NewProcess creates a process
func NewProcess(
	actorStorage storage.ActorStorage,
	emitterID,
	envelopeID api.UUID,
	graph *graph.Graph,
	source storage.NodeOutputRef,
	manager *Manager,
) *Process {
	id, err := api.UUIDV4()
	if err != nil {
		panic(fmt.Sprintf("Cannot generate a UUID: %s", err))
	}
	return &Process{
		ID:                id,
		actors:            actorStorage,
		SourceRef:         source,
		TriggerEmitterID:  emitterID,
		Graph:             graph,
		TriggerEnvelopeID: envelopeID,
		manager:           manager,
	}
}

// Process orchestrate a graph execution
type Process struct {
	ID api.UUID

	actors         storage.ActorStorage
	processStorage storage.ProcessStorage
	graphManager   *graph.Manager

	manager *Manager

	TriggerEmitterID  api.UUID
	TriggerEnvelopeID api.UUID

	Graph     *graph.Graph
	SourceRef storage.NodeOutputRef
}

// GetStatus returns the current status of the process
func (p Process) GetStatus() api.Process_Status {
	state := p.processStorage.GetState(p.ID)
	return state.Status
}

// Pause switch the process state to ProcessPaused
func (p Process) Pause() error {
	return p.processStorage.SetStatus(p.ID, api.Process_PAUSED)
}

// Resume switch the process state to ProcessRunning if previously paused
func (p Process) Resume() error {
	return p.processStorage.SetStatus(p.ID, api.Process_RUNNING)
}

// MatchGraph attempts to find the graph that will drive the process
func (p *Process) MatchGraph(eventTypes []string) error {
	var success bool
	// TODO UpdateReceiveStatus with ERROR if no match
	defer func() {
		if !success {
			p.manager.Log.Errorf("MatchGraph failed, setting status to error")
			if err := p.processStorage.SetStatus(p.ID, api.Process_ERROR); err != nil {
				p.manager.Log.Errorf("Error setting process status: %s", err)
			}
		}
	}()
	actor, ok := p.actors.Get(p.TriggerEmitterID)
	if !ok {
		return fmt.Errorf("Invalid actor")
	}
	g, outputRef, ok := p.graphManager.LookupGraph(actor, eventTypes)
	if !ok {
		return fmt.Errorf("No matching graph")
	}
	p.Graph = &g
	p.SourceRef = outputRef
	p.processStorage.SetGraph(p.ID, p.Graph.ID, p.SourceRef)
	success = true
	return nil
}

// GetTargets returns the targets for an envelope coming from a given emitter
func (p Process) GetTargets(output storage.NodeOutputRef, envelopeID api.UUID, noRouteTableUpdate bool) ([]api.EnvelopeTarget, error) {
	if p.Graph == nil {
		return nil, fmt.Errorf("Graph-less process")
	}
	// is the envelope already routed ?
	return p.processStorage.GetTargets(p.ID, output, envelopeID, noRouteTableUpdate,
		func() ([]api.EnvelopeTarget, error) {

			dests, err := p.Graph.GetDestinations(output)
			if err != nil {
				return nil, err
			}
			if len(dests) != 1 {
				panic("can only handle a single destination")
			}
			dest := dests[0]
			destNode, err := p.Graph.GetNode(dest.NodeID)
			if err != nil {
				return nil, fmt.Errorf("Cannot find the target node: %s", err)
			}

			var actors []api.Actor

			if destNode.Type == storage.ConsumerNode {
				if !destNode.RoleBroadcast || len(destNode.Roles) != 1 {
					panic("not implemented")
				}

				actors = p.actors.ListRole(api.Actor_CONSUMER, destNode.Roles[0])
			} else if destNode.Type == storage.WorkerNode {
				if destNode.RoleBroadcast || len(destNode.Roles) != 0 {
					panic("Worker graph nodes supports only Actors, not roles")
				}

				for _, actorID := range destNode.ActorIDs {
					actor, exists := p.actors.Get(actorID)
					if !exists {
						return nil, fmt.Errorf("No such actor: %s", actorID)
					}
					actors = append(actors, actor)
				}

				for _, actorRef := range destNode.Actors {
					actorID := api.MaybeUUIDFromString(actorRef)
					if !actorID.IsZero() {
						actor, exists := p.actors.Get(actorID)
						if !exists {
							return nil, fmt.Errorf("No such actor: %s", actorID)
						}
						actors = append(actors, actor)
					} else {
						actor, exists := p.actors.GetByName(actorRef)
						if !exists {
							return nil, fmt.Errorf("No such actor: %s", actorRef)
						}
						actors = append(actors, actor)
					}
				}
			}

			if len(actors) == 0 {
				return nil, fmt.Errorf("No target actor found")
			}
			var targets []api.EnvelopeTarget
			for _, actor := range actors {
				targets = append(targets, api.EnvelopeTarget{
					ActorID: actor.ID,
					NodeID:  dest.NodeID,
					Input:   dest.Input,
				})
			}
			return targets, nil
		},
	)

}

// UpdateReceiveStatus is called whenever an envelope fragment is received from a node
func (p *Process) UpdateReceiveStatus(envelopeID, actorID api.UUID, outputref storage.NodeOutputRef, status storage.ReceptionStatus) error {
	return p.processStorage.UpdateReceiveStatus(
		storage.Reception{
			ProcessID:       p.ID,
			EnvelopeID:      envelopeID,
			SourceActor:     actorID,
			SourceOutputRef: outputref,
		},
		status)
}

// ActorProcessingEnd informs the process
func (p *Process) ActorProcessingEnd(actorID api.UUID, nodeID string, status api.ActorProcessingState_Status) {
	p.processStorage.ProcessNodeEnd(
		p.ID, storage.ProcessNode{ActorID: actorID, NodeID: nodeID}, status)
}

// GetPostMortemState set the postmortem state of the process
func (p *Process) GetPostMortemState() (storage.PostMortemState, error) {
	if status := p.GetStatus(); status != api.Process_DONE && status != api.Process_ERROR {
		return storage.PostMortemState{}, fmt.Errorf("Cannot get postmortem state on a non-ended process. Current status is '%s'", status)
	}
	return p.processStorage.GetPostMortemState(p.ID), nil
}

// SetPostMortemState set the postmortem state of the process
func (p *Process) SetPostMortemState(state storage.PostMortemState) error {
	if status := p.GetStatus(); status != api.Process_DONE && status != api.Process_ERROR {
		return fmt.Errorf("Cannot set postmortem state on a non-ended process. Current status is '%s'", status)
	}
	p.processStorage.SetPostMortemState(p.ID, state)
	return nil
}

func indent(buf, pad []byte) (out []byte) {
	lines := bytes.Split(buf, []byte("\n"))
	for _, line := range lines {
		out = append(out, pad...)
		out = append(out, line...)
		out = append(out, '\n')
	}
	return
}

// ExportStream export the process as a stream, which supports big envelopes
// and logs
func (p *Process) ExportStream(
	ctx context.Context,
	envelopeStorage storage.EnvelopeStorage,
	log xbus.Logger,
	out chan<- []byte,
	exportEnvelope bool,
	envelopeSizeLimit int64,
) error {

	var export = make(map[string]interface{})
	export["id"] = p.ID

	emitterEx := map[string]interface{}{
		"id": p.TriggerEmitterID,
	}

	if emitter, exists := p.actors.Get(p.TriggerEmitterID); exists {
		emitterEx["name"] = emitter.Name
	}

	export["trigger"] = map[string]interface{}{
		"emitter":  emitterEx,
		"envelope": p.TriggerEnvelopeID,
	}

	if p.Graph != nil {
		export["graph"] = map[string]interface{}{
			"id":      p.Graph.ID,
			"name":    p.Graph.Name,
			"version": p.Graph.Version,
			"status":  p.Graph.Status,
			"nodes":   p.Graph.Nodes,
			"edges":   p.Graph.Edges,
		}
	} else {
		export["graph"] = nil
	}

	export["status"] = p.GetStatus()
	if p.GetStatus() == api.Process_DONE || p.GetStatus() == api.Process_ERROR {
		pmState, err := p.GetPostMortemState()
		if err != nil {
			panic(err)
		}
		export["postmortem_state"] = pmState
	}

	export["state"] = p.processStorage.GetState(p.ID)

	b, err := yaml.Marshal([]interface{}{export})
	if err != nil {
		return nil
	}

	out <- b

	if exportEnvelope {
		out <- []byte("  envelopes:\n")
		envelopeIDs := []api.UUID{p.TriggerEnvelopeID}
		var size int64

		for _, id := range envelopeIDs {
			if ctx.Err() != nil {
				return nil
			}
			envelope, pos, err := envelopeStorage.ReadEnvelope(
				id, storage.EnvelopeStartPos(), envelopeExportChunkSize)
			if err != nil {
				log.Errorf("Error reading envelope %s: %s", id, err)
				continue
			}
			size += envelopeExportChunkSize
			var b []byte
			if pos.Complete {
				b, err = yaml.Marshal(map[string]interface{}{
					id.String(): envelope,
				})
			} else {
				b, err = yaml.Marshal(map[string][]interface{}{
					id.String(): {envelope},
				})
			}
			if err != nil {
				panic(err)
			}
			if ctx.Err() != nil {
				return nil
			}
			out <- indent(b, []byte("    "))

			lastPos := pos
			for !pos.Complete && (envelopeSizeLimit == 0 || size < envelopeSizeLimit) {
				envelope, pos, err = envelopeStorage.ReadEnvelope(
					id, lastPos, envelopeExportChunkSize,
				)
				if err != nil {
					log.Errorf("Error reading envelope %s: %s", id, err)
					break
				}
				size += envelopeExportChunkSize
				if pos.Equals(lastPos) {
					log.Errorf("Cannot get more of the envelope")
					break
				}
				b, err = yaml.Marshal([]interface{}{envelope})
				if err != nil {
					panic(err)
				}
				if ctx.Err() != nil {
					return nil
				}
				out <- indent(b, []byte("    "))
				lastPos = pos
			}
		}
	}

	return nil
}

// Export exports all the data of the process, including envelopes and logs,
// as a map suitable for json or yaml export
// This implementation is not big-envelope friendly, a streamed version will
// replace it in the future
func (p *Process) Export(envelopeStorage storage.EnvelopeStorage, log xbus.Logger) (map[string]interface{}, error) {
	var export = make(map[string]interface{})
	export["id"] = p.ID

	emitterEx := map[string]interface{}{
		"id": p.TriggerEmitterID,
	}

	if emitter, exists := p.actors.Get(p.TriggerEmitterID); exists {
		emitterEx["name"] = emitter.Name
	}

	export["trigger"] = map[string]interface{}{
		"emitter":  emitterEx,
		"envelope": p.TriggerEnvelopeID,
	}

	if p.Graph != nil {
		export["graph"] = map[string]interface{}{
			"id":      p.Graph.ID,
			"name":    p.Graph.Name,
			"version": p.Graph.Version,
			"status":  p.Graph.Status,
			"nodes":   p.Graph.Nodes,
			"edges":   p.Graph.Edges,
		}
	} else {
		export["graph"] = nil
	}

	export["status"] = p.GetStatus()
	if p.GetStatus() == api.Process_DONE || p.GetStatus() == api.Process_ERROR {
		pmState, err := p.GetPostMortemState()
		if err != nil {
			panic(err)
		}
		export["postmortem_state"] = pmState
	}

	export["state"] = p.processStorage.GetState(p.ID)

	envelopeIDs := []api.UUID{p.TriggerEnvelopeID}
	var envelopes []api.Envelope

	for _, id := range envelopeIDs {
		envelope, pos, err := envelopeStorage.ReadEnvelope(
			id, storage.EnvelopeStartPos(), 100*1024*1024)
		if err != nil {
			log.Errorf("Error reading envelope %s: %s", id, err)
			continue
		}
		if !pos.Complete {
			log.Errorf("Envelope %s is too big (>100Ko)", id)
			continue
		}
		envelopes = append(envelopes, envelope)
	}

	export["envelopes"] = envelopes

	return export, nil
}
