// Copyright 2016 Orus.io. All rights reserved.

package sql

import (
	"github.com/orus-io/yago"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"github.com/m4rw3r/uuid"
)

// NewAccountStorage initialize a new AccountStorage
func NewAccountStorage(model *model.Model, db *yago.DB, logger xbus.Logger) *AccountStorage {
	return &AccountStorage{Base{
		model:  model,
		db:     db,
		logger: logger,
	}}
}

// AccountStorage implements all the persistor interfaces
type AccountStorage struct {
	Base
}

func toDB(sAccount api.Account) model.Account {
	return model.Account{
		Base:   model.Base{ID: uuid.UUID(sAccount.ID)},
		Name:   sAccount.Name,
		Type:   sAccount.Type,
		Status: sAccount.Status,
		CSR:    sAccount.CSR,
		Cert:   sAccount.Cert,
		APIKey: sAccount.APIKey,
	}
}

func fromDB(dbAccount model.Account) api.Account {
	return api.Account{
		ID:     api.UUID(dbAccount.ID),
		Name:   dbAccount.Name,
		Type:   dbAccount.Type,
		Status: dbAccount.Status,
		CSR:    dbAccount.CSR,
		Cert:   dbAccount.Cert,
		APIKey: dbAccount.APIKey,
	}
}

// Get returns an account from its ID
func (p AccountStorage) Get(id api.UUID) (api.Account, bool) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	var dbAccount model.Account
	err := tx.Query(p.model.Account).Get(&dbAccount, uuid.UUID(id))
	if err == yago.ErrRecordNotFound {
		return api.Account{}, false
	}
	if err != nil {
		panic(err)
	}
	return fromDB(dbAccount), true
}

// GetByName returns an account with the given name if exists
func (p AccountStorage) GetByName(name string) (api.Account, bool) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	var dbAccount model.Account
	err := tx.Query(p.model.Account).Where(
		p.model.Account.Name.Eq(name)).
		One(&dbAccount)
	if err == yago.ErrRecordNotFound {
		return api.Account{}, false
	}
	if err != nil {
		panic(err)
	}
	return fromDB(dbAccount), true
}

// GetByCert returns an account with the given cert if exists
func (p AccountStorage) GetByCert(cert string) (account api.Account, exists bool) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	var dbAccount model.Account
	err := tx.Query(p.model.Account).Where(
		p.model.Account.Cert.Eq(cert)).
		One(&dbAccount)
	if err == yago.ErrRecordNotFound {
		return api.Account{}, false
	}
	if err != nil {
		panic(err)
	}
	return fromDB(dbAccount), true
}

// GetByAPIKey returns an account with the given api key if exists
func (p AccountStorage) GetByAPIKey(key string) (account api.Account, exists bool) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	var dbAccount model.Account
	err := tx.Query(p.model.Account).Where(
		p.model.Account.APIKey.Eq(key)).
		One(&dbAccount)
	if err == yago.ErrRecordNotFound {
		return api.Account{}, false
	}
	if err != nil {
		panic(err)
	}
	return fromDB(dbAccount), true
}

// Create creates a new account
func (p AccountStorage) Create(account api.Account) (api.Account, error) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	dbAccount := toDB(account)
	err := tx.Insert(&dbAccount)
	if err != nil {
		// TODO on constraint error, return a proper error and don't panic
		panic(err)
	}
	tx.MustCommit()

	return fromDB(dbAccount), nil
}

// Delete deletes an account
func (p AccountStorage) Delete(account api.Account) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	dbAccount := toDB(account)
	err := tx.Delete(&dbAccount)
	if err != nil {
		panic(err)
	}
	tx.MustCommit()
}

// Update updates an account
func (p AccountStorage) Update(account api.Account) (api.Account, error) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	dbAccount := toDB(account)
	err := tx.Update(&dbAccount)
	if err != nil {
		// TODO on constraint error, return a proper error and don't panic
		panic(err)
	}
	tx.MustCommit()
	return fromDB(dbAccount), nil
}

// List returns a list of all the accounts
func (p AccountStorage) List() []api.Account {
	var (
		dbAccountList []model.Account
	)
	tx := p.MustBegin()
	defer tx.Rollback()

	must(tx.Query(p.model.Account).All(&dbAccountList))

	var res []api.Account
	for _, a := range dbAccountList {
		res = append(res, api.Account{
			ID:     api.UUID(a.ID),
			Name:   a.Name,
			Type:   a.Type,
			Status: a.Status,
			CSR:    a.CSR,
			Cert:   a.Cert,
		})
	}
	return res
}
