// Copyright 2017 Orus.io. All rights reserved.

package uuid

import (
	"bytes"
	"database/sql/driver"
	"reflect"

	muuid "github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
)

// NullUUID is a api.UUID that handles IsZero as a db NULL value,
// and as an empty string (or null) in (un)marshalText
type NullUUID api.UUID

// FromString initialize a NullUUID from a string
func FromString(str string) (NullUUID, error) {
	var u NullUUID
	return u, u.SetString(str)
}

// MustFromString initialize a NullUUID from a string and panics if
// it cannot
func MustFromString(str string) NullUUID {
	u, err := FromString(str)
	if err != nil {
		panic(err)
	}
	return u
}

// FromBytes initialize a NullUUID from []byte
func FromBytes(str []byte) (NullUUID, error) {
	var u NullUUID
	return u, u.ReadBytes(str)
}

// MustFromBytes initialize a NullUUID from a []byte and panics if
// it cannot
func MustFromBytes(str []byte) NullUUID {
	u, err := FromBytes(str)
	if err != nil {
		panic(err)
	}
	return u
}

// IsZero returns true if value is zero
func (u NullUUID) IsZero() bool {
	return api.UUID(u).IsZero()
}

// IsNull returns true if value is Zero
func (u NullUUID) IsNull() bool {
	return u.IsZero()
}

// SetZero reset the value to empty
func (u *NullUUID) SetZero() {
	(*api.UUID)(u).SetZero()
}

// SetNull reset the value to empty
func (u *NullUUID) SetNull() {
	u.SetZero()
}

func (u NullUUID) String() string {
	if u.IsNull() {
		return ""
	}
	return api.UUID(u).String()
}

// SetString see api.UUID.SetString
func (u *NullUUID) SetString(str string) error {
	if str == "" || str == "null" {
		u.SetNull()
		return nil
	}
	return (*api.UUID)(u).SetString(str)
}

// ReadBytes see api.UUID.ReadBytes
func (u *NullUUID) ReadBytes(str []byte) error {
	if len(str) == 0 || bytes.Equal([]byte("null"), str) {
		u.SetNull()
		return nil
	}
	return (*api.UUID)(u).ReadBytes(str)
}

// Scan scans a uuid from the given interface instance.
func (u *NullUUID) Scan(val interface{}) error {
	if val == nil {
		*u = NullUUID{}
		return nil
	}
	if s, ok := val.(string); ok {
		return u.SetString(s)
	}
	if b, ok := val.([]byte); ok {
		return u.ReadBytes(b)
	}
	return &muuid.ErrInvalidType{Type: reflect.TypeOf(val)}
}

// Value gives the database driver representation of the UUID.
func (u NullUUID) Value() (driver.Value, error) {
	if u.IsZero() {
		return nil, nil
	}
	return u.String(), nil
}

// MarshalText returns the stringified UUID, or an empty slice if the value is Null/Zero
func (u NullUUID) MarshalText() ([]byte, error) {
	return []byte(u.String()), nil
}

// UnmarshalText returns the stringified UUID, or nil if the value is Null/Zero
func (u *NullUUID) UnmarshalText(text []byte) error {
	return u.ReadBytes(text)
}
